package client;

import java.awt.Color;
import java.awt.Rectangle;
import java.awt.Robot;
import java.awt.image.BufferedImage;
import java.io.File;
import java.util.concurrent.atomic.AtomicBoolean;

import javax.imageio.ImageIO;
import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.Clip;

import org.lwjgl.LWJGLException;
import org.lwjgl.input.Mouse;
import org.lwjgl.opengl.Display;
import org.lwjgl.opengl.DisplayMode;

import client.model.Client;

import com.threed.jpct.Config;
import com.threed.jpct.FrameBuffer;
import com.threed.jpct.IRenderer;
import com.threed.jpct.Logger;

public class Engine implements Runnable {

	private final int FRAME_WIDTH = 800;
	private final int FRAME_HEIGHT = 600;

	private final long TIMER_RESOLUTION = 1000 * 1000 * 1000;
	private final int DELAYS_PER_YIELD = 10;
	private final int MAX_FRAME_SKIPS = 10;
	private long totalUpdateCount = 0L;
	private float lastUpdateCount = 0L;
	private long secondsRunning = 0L;
	private final int requestedFPS = 30;
	private long lastFrameCount = 0L;
	@SuppressWarnings("unused")
	private long averageFrameCount = 0L;
	private static AtomicBoolean running = new AtomicBoolean(true);
	private static final String FILE_STORAGE = getFileStorage(false, "aeonlite");

	private final FrameBuffer frameBuffer;
	private final Client client;

	public static void main(String[] args) {
		//System.setProperty("sun.java2d.noddraw", "true");
		/** 
		 * Implement the 3D native library.
		 */
		boolean windows = System.getProperty("os.name").toLowerCase().contains("win");
		System.setProperty("org.lwjgl.librarypath", new File((windows ? "./file_storage/" : FILE_STORAGE) + "java" + File.separator + (windows ? "windows" : "linux") + File.separator).getAbsolutePath());
		new Thread(new Engine()).run();
	}

	public Engine() {
		/**
		 * Configure jPCT
		 */
		Config.glVerbose = true;
		Logger.setLogLevel(Logger.LL_ONLY_ERRORS);
		Config.glColorDepth = 24;
		Config.autoBuild = true;
		Config.glAvoidTextureCopies = true;
		Config.oldStyle3DSLoader = true;
		//Config.useMultipleThreads = Runtime.getRuntime().availableProcessors() > 1;

		Config.glTrilinear = !Settings.PERFORMANCE_MODE; // Default is false.
		// Lets jPCT generate Mipmaps for all textures and enables trilinear filtering between them.
		// This has to be set before enabling the OpenGL renderer to have an effect.

		Config.texelFilter = !Settings.PERFORMANCE_MODE;
		// Use texel-filtering to avoid the blocky look of point sampling.
		// Texel filtering does cost some performance (but not much) and may introduce artifacts in some cases.
		// Filtering is only applied to texels that are "in need of" and ignored when using OpenGL.

		Config.glUseFBO = true; // Default is true.
		// Makes jPCT use frame buffer objects (FBOs) if possible and applicable.
		// In case of compability problems with render to texture and/or shadow mapping, try setting this to false.
		// It will hurt performance though. Set this before enabling a renderer.
		// NOTE: This value will be adjusted by jPCT after initializing an OpenGL-renderer to the actually used state.

		Config.nativeBufferSize = Settings.PERFORMANCE_MODE ? 512 : 1024; // Default is 1024.
		// The size in kb of the native buffer cache.
		// Setting this to 0 will turn native buffer caching off.
		// Setting this to a very high value may lead, depending on the application, to a kind of memory leak.

		Config.useBB = true; // Default is true.
		// Use the bounding box every Object3D has for faster clipping and culling.

		/**
		 * Crashes client..
		int numberOfProcs = Runtime.getRuntime().availableProcessors();
		Config.useMultipleThreads = numberOfProcs > 1;
		Config.useMultiThreadedBlitting = numberOfProcs > 1;
		Config.loadBalancingStrategy = 1;
		Config.maxNumberOfCores = numberOfProcs;
		 */

		/**
		 * Initialize the jPCT application
		 */
		try {
			//Config.glFullscreen = true;
			Config.glSkipInitialization = true;
			Config.glWindowName = "Aeonlite";
			Config.glVSync = false; // If set to true, vertical sync will be used if the driver supports this setting and doesn't override it itself.
			if (!Settings.FULL_SCREEN) {
				Display.setDisplayMode(new DisplayMode(FRAME_WIDTH, FRAME_HEIGHT));
			}
			Display.setTitle(Config.glWindowName);
			Display.sync(60);
			Display.setResizable(false);
			Display.setFullscreen(Settings.FULL_SCREEN);
			Display.create();
		} catch (LWJGLException e) {
			System.err.println("Fatal error creating display! Exception: "+e.toString());
			System.exit(1);
		}

		this.client = new Client(FILE_STORAGE);
		this.frameBuffer = new FrameBuffer(Display.getWidth(), Display.getHeight(), FrameBuffer.SAMPLINGMODE_NORMAL);
		//this.frameBuffer = new FrameBuffer(full_screen ? Display.getWidth() : FRAME_WIDTH, full_screen ? Display.getHeight() : FRAME_HEIGHT, FrameBuffer.SAMPLINGMODE_NORMAL);
		frameBuffer.enableRenderer(IRenderer.RENDERER_OPENGL, IRenderer.MODE_OPENGL);
		frameBuffer.disableRenderer(IRenderer.RENDERER_SOFTWARE);
	}

	@Override
	public void run() {
		long startTime = 0L;
		long endTime = 0L;
		long timeDiff = 0L;
		long sleepTime = 0L;
		long overSleepTime = 0L;
		long excess = 0L;
		int numberOfDelays = 0;
		int frameSkips = 0;
		int frameCount = 0;
		long period = TIMER_RESOLUTION / requestedFPS;
		long totalFrameCount = 0L;
		float currentUPS = 0;
		float delta = 0;
		long previousSampleTime = System.nanoTime();
		while (running.get() && !Display.isCloseRequested()) {
			startTime = System.nanoTime(); 
			if (Mouse.isCreated() && frameBuffer.isInitialized() && Display.isCreated()) {
				if (Display.wasResized()) {
					frameBuffer.resize(Display.getWidth(), Display.getHeight());
				}

				// Update loop
				currentUPS = totalUpdateCount / (float) (secondsRunning == 0 ? 1 : secondsRunning);
				delta = requestedFPS / currentUPS;
				update(delta);

				// Rendering
				frameBuffer.clear(Color.BLACK); // Clear last frame
				// Render new frame
				Config.glTransparencyOffset = 0.0f;
				Config.glTransparencyMul = 0.025f;
				client.render3D(frameBuffer); // 3D first
				Config.glTransparencyOffset = 0.7f;
				Config.glTransparencyMul = 0.06f;
				client.render2D(frameBuffer, lastFrameCount, (int) lastUpdateCount); // 2D second
				frameBuffer.update();
				frameBuffer.displayGLOnly(); // End of cycle
			}
			endTime = System.nanoTime();
			timeDiff = endTime - startTime;
			sleepTime = (period - timeDiff) - overSleepTime;
			if (sleepTime > 0) {
				try {
					Thread.sleep(sleepTime / (TIMER_RESOLUTION / 1000));
				} catch(final InterruptedException ignore) {
					Thread.currentThread().interrupt();
					break;
				}
				overSleepTime = (System.nanoTime() - endTime) - sleepTime;
			} else {
				excess -= sleepTime;
				overSleepTime = 0L;
				numberOfDelays++;
				if (numberOfDelays > DELAYS_PER_YIELD) {
					Thread.yield();
					numberOfDelays = 0;
				}
			}
			frameSkips = 0;
			while ((excess > period) && (frameSkips < MAX_FRAME_SKIPS)) {
				excess -= period;
				update(delta);
				frameSkips++;
			}
			frameCount++;
			totalFrameCount++;
			totalUpdateCount += frameSkips + 1;
			if (System.nanoTime() - previousSampleTime > TIMER_RESOLUTION) {
				secondsRunning++;
				lastFrameCount = frameCount;
				lastUpdateCount = currentUPS;
				frameCount = 0;
				averageFrameCount = totalFrameCount / secondsRunning;
				previousSampleTime =  System.nanoTime();
			}
		}
		Mouse.destroy();
		frameBuffer.dispose();
		Display.destroy();
		Runtime.getRuntime().runFinalization();
		Runtime.getRuntime().gc();
		Runtime.getRuntime().exit(0);
	}

	private void update(float delta) {
		if (delta <= 0) {
			return;
		}
		client.update(delta, frameBuffer);
	}

	private static String getFileStorage(boolean published, String title) {
		if (published) {
			String path = ".dse_console" + File.separator + "games" + File.separator + title + File.separator;
			return System.getProperty("user.home", ".") + path;
		} else {
			return "./file_storage/";
		}
	}

	public static int getMouseX() {
		return Mouse.getX();
	}

	public static int getMouseY() {
		return Display.getHeight() - Mouse.getY();//- (Mouse.getY() - mouseYOffset);
	}

	/**
	 * A static method that will play an audio file.
	 */
	public static synchronized void playAudio(final String wavFile) {
		new Thread(new Runnable() {
			// The wrapper thread is unnecessary, unless it blocks on the Clip finishing..
			public void run() {
				try {
					Clip clip = AudioSystem.getClip();
					//AudioInputStream inputStream = AudioSystem.getAudioInputStream(Engine.class.getResourceAsStream("/path/to/sounds/" + wavFile));
					AudioInputStream inputStream = AudioSystem.getAudioInputStream(new File(FILE_STORAGE + "audio" + File.separator + wavFile + ".wav"));
					clip.open(inputStream);
					clip.start(); 
				} catch (Exception e) {
					System.err.println(e.getMessage());
				}
			}
		}).start();
	}

	/**
	 * A static method that will save a screen shot.
	 */
	public static void screenShot() {
		String prefix = "./cgi" + File.separator;
		try {
			int x = 0;
			int y = 0;
			int w = Display.getWidth();
			int h = Display.getHeight();
			Robot robot = new Robot();
			Rectangle captureSize = new Rectangle(x, y, w, h);
			BufferedImage bufferedimage = robot.createScreenCapture(captureSize);
			long picNumber = System.nanoTime();
			File file = new File((new StringBuilder()).append(prefix + "ss").append(picNumber).append(".png").toString());
			ImageIO.write(bufferedimage, "png", file);
			System.out.println("A moment in time has been captured.");
		} catch(Exception e) {
			e.printStackTrace();
		}
	}

	/**
	 * A static method will terminate the application.
	 */
	public static void terminate() {
		System.out.println("Client terminated by user.");
		running.set(false);
	}

}