package client.model;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.StreamTokenizer;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.concurrent.atomic.AtomicInteger;

import client.Settings;
import client.util.Utils;

import com.threed.jpct.Camera;
import com.threed.jpct.Config;
import com.threed.jpct.Loader;
import com.threed.jpct.Object3D;
import com.threed.jpct.SimpleVector;
import com.threed.jpct.Texture;
import com.threed.jpct.TextureManager;
import com.threed.jpct.World;

public class NpcManager {

	// Model cache
	private final Map<Integer, Object3D> models = new HashMap<Integer, Object3D>();

	// Dynamic npcs
	private final List<Npc> npcs = new CopyOnWriteArrayList<Npc>();
	private AtomicInteger uid = new AtomicInteger(1);
	private long lastVisibilityUpdate = 0L;

	private final Object3D warrior; // (paid)
	private final Object3D archer; // (paid)
	private final Object3D wizard; // (paid)
	private final Object3D dragon; // (paid)
	private final Object3D zombie; // (paid)
	private final Object3D ape; // (free)
	private final Object3D goblin; // (free)
	private final Object3D skeleton; // (free)
	private final Object3D protector; // (free)

	private final TextureManager TM = TextureManager.getInstance();
	
	public NpcManager(final String fileStorage, final World world, final WorldMap terrain) {
		Texture t; // recycled
		
		this.warrior = Loader.loadMD2(fileStorage + "npc" + File.separator + "warrior.md2", 0.9f);
		TM.addTexture("warrior", new Texture(fileStorage + "npc" + File.separator + "warrior.jpg"));
		t = TM.getTexture("warrior");
		t.setMipmap(!Settings.PERFORMANCE_MODE);
		warrior.setTexture("warrior");
		warrior.rotateY((float) Math.PI * 1.5f); // Properly rotate the MD2 model
		warrior.rotateMesh();
		warrior.getAnimationSequence().remove(2); // death 2
		warrior.getAnimationSequence().remove(5); // run
		warrior.getAnimationSequence().remove(7); // kick
		warrior.compile(true); // Compile dynamic object
		models.put(0, warrior);

		this.archer = Loader.loadMD2(fileStorage + "npc" + File.separator + "archer.md2", 0.9f);
		TM.addTexture("archer", new Texture(fileStorage + "npc" + File.separator + "archer.jpg"));
		t = TM.getTexture("archer");
		t.setMipmap(!Settings.PERFORMANCE_MODE);
		archer.setTexture("archer");
		archer.rotateY((float) Math.PI * 1.5f); // Properly rotate the MD2 model
		archer.rotateMesh();
		archer.getAnimationSequence().remove(2); // death 2
		archer.getAnimationSequence().remove(5); // run
		archer.getAnimationSequence().remove(7); // kick
		archer.compile(true); // Compile dynamic object
		models.put(1, archer);

		this.wizard = Loader.loadMD2(fileStorage + "npc" + File.separator + "wizard.md2", 0.9f);
		TM.addTexture("wizard", new Texture(fileStorage + "npc" + File.separator + "wizard.jpg"));
		t = TM.getTexture("wizard");
		t.setMipmap(!Settings.PERFORMANCE_MODE);
		wizard.setTexture("wizard");
		wizard.rotateY((float) Math.PI * 1.5f); // Properly rotate the MD2 model
		wizard.rotateMesh();
		wizard.getAnimationSequence().remove(2); // death 2
		wizard.getAnimationSequence().remove(5); // run
		wizard.getAnimationSequence().remove(7); // kick
		wizard.compile(true); // Compile dynamic object
		models.put(2, wizard);

		this.dragon = Loader.loadMD2(fileStorage + "npc" + File.separator + "dragon.md2", .13f);
		TM.addTexture("dragon", new Texture(fileStorage + "npc" + File.separator + "dragon.jpg"));
		t = TM.getTexture("dragon");
		t.setMipmap(!Settings.PERFORMANCE_MODE);
		dragon.setTexture("dragon");
		dragon.rotateY((float) Math.PI * 1.5f); // Properly rotate the MD2 model
		dragon.rotateMesh();
		dragon.getAnimationSequence().remove(7); // remove fly animation
		dragon.compile(true); // Compile dynamic object
		models.put(3, dragon);

		this.ape = Loader.loadMD2(fileStorage + "npc" + File.separator + "ape.md2", .20f);
		TM.addTexture("ape", new Texture(fileStorage + "npc" + File.separator + "ape.jpg"));
		t = TM.getTexture("ape");
		t.setMipmap(!Settings.PERFORMANCE_MODE);
		ape.setTexture("ape");
		ape.rotateY((float) Math.PI * 1.5f); // Properly rotate the MD2 model
		ape.rotateMesh();
		ape.compile(true); // Compile dynamic object
		models.put(4, ape);

		TM.addTexture("zombie", new Texture(fileStorage + "npc" + File.separator + "zombie.jpg"));
		this.zombie = Loader.loadMD2(fileStorage + "player" + File.separator + "male.md2", 0.10f);
		t = TM.getTexture("zombie");
		t.setMipmap(!Settings.PERFORMANCE_MODE);
		zombie.setTexture("zombie");
		zombie.rotateY((float) Math.PI * 1.5f); // Properly rotate the MD2 model
		zombie.rotateMesh();
		zombie.compile(true); // Compile dynamic object
		models.put(5, zombie);

		this.goblin = Loader.loadMD2(fileStorage + "npc" + File.separator + "goblin.md2", 0.24f);
		TM.addTexture("goblin", new Texture(fileStorage + "npc" + File.separator + "goblin.png"));
		t = TM.getTexture("goblin");
		t.setMipmap(!Settings.PERFORMANCE_MODE);
		goblin.setTexture("goblin");
		goblin.rotateY((float) Math.PI * 1.5f); // Properly rotate the MD2 model
		goblin.rotateMesh();
		for (int i = 4; i < 16; i++) {
			goblin.getAnimationSequence().remove(i); // unused animations..
		}
		goblin.compile(true); // Compile dynamic object
		models.put(6, goblin);
		
		this.skeleton = Loader.loadMD2(fileStorage + "npc" + File.separator + "skeleton.md2", 0.36f);
		TM.addTexture("skeleton", new Texture(fileStorage + "npc" + File.separator + "skeleton.jpg"));
		t = TM.getTexture("skeleton");
		t.setMipmap(!Settings.PERFORMANCE_MODE);
		skeleton.setTexture("skeleton");
		skeleton.rotateY((float) Math.PI * 1.5f); // Properly rotate the MD2 model
		skeleton.rotateMesh();
		skeleton.compile(true); // Compile dynamic object
		models.put(7, skeleton);

		this.protector = Loader.loadMD2(fileStorage + "npc" + File.separator + "protector.md2", 0.38f);
		TM.addTexture("protector", new Texture(fileStorage + "npc" + File.separator + "protector.jpg"));
		t = TM.getTexture("protector");
		t.setMipmap(!Settings.PERFORMANCE_MODE);
		protector.setTexture("protector");
		protector.rotateY((float) Math.PI * 1.5f); // Properly rotate the MD2 model
		protector.rotateMesh();
		protector.compile(true); // Compile dynamic object
		models.put(8, protector);

		for (int i = 0; i < models.size(); i++) {
			models.get(i).build();
		}

		/**
		 * Read the npc configuration file.
		 */
		try {
			File file = new File(fileStorage + "config" + File.separator + "npcs.cfg");
			if (!file.exists()) {
				System.out.println("Could not load: " + file.getAbsolutePath());
			}
			StreamTokenizer tokenizer = new StreamTokenizer(new BufferedReader(new FileReader(file)));
			while (tokenizer.nextToken() != StreamTokenizer.TT_EOF) {
				int type = (int) tokenizer.nval;
				tokenizer.nextToken();
				tokenizer.nextToken();
				int x = (int) tokenizer.nval;
				tokenizer.nextToken();
				tokenizer.nextToken();
				int y = (int) tokenizer.nval;
				tokenizer.nextToken();
				tokenizer.nextToken();
				int z = (int) tokenizer.nval;
				Npc npc = new Npc(getModelByType(type), type, uid.getAndIncrement(), new SimpleVector(x, y, z), world, terrain); // clone the model
				npcs.add(npc);
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
		System.out.println("Loaded " + models.size() + " npc models, and " + npcs.size() + " spawns.");
	}
	
	/**
	 * Updates the visibility list every 6 seconds (10x per minute)
	 */
	private int numberOfVisibleNpcs = 0;
	public void updateVisibilityList(final Camera camera, final boolean forceUpdate) {
		if (forceUpdate || lastVisibilityUpdate <= System.currentTimeMillis() - Settings.ENTITY_VISIBILITY_UPDATE_DELAY) {
			lastVisibilityUpdate = System.currentTimeMillis();
			numberOfVisibleNpcs = 0;
			for (Npc npc : npcs) {
				if (!Utils.isInRange(Config.farPlane * 1.25f, camera.getPosition(), npc.getTransformedCenter())) {
					npc.setVisibility(false);
				} else {
					npc.setVisibility(true);
					numberOfVisibleNpcs++;
				}
			}
		}
	}

	public Npc getInteractionEntity(final Object3D focusPoint) {
		for (Npc npc : npcs) {
			if (npc.getVisibility() && Utils.isInRange(20f, focusPoint, npc)) {
				return npc;
			}
		}
		return null;
	}
	
	/**
	 * Cycle through the prop list, and update each one.
	 * @note Processes on a 36ms delay
	 */
	private long lastCycle = 0L;
	public void cycle(final float delta) {
		if (lastCycle <= System.currentTimeMillis() - 36) {
			lastCycle = System.currentTimeMillis();
			for (Npc npc : npcs) {
				npc.update(delta);
			}
		}
	}
	
	public List<Npc> getNpcList() {
		return npcs;
	}

	public Object3D getModelByType(int type) {
		if (!models.containsKey(type)) {
			System.out.println("Npc (type: "+type+") does not exist!!");
			return null;
		}
		return models.get(type).cloneObject();
	}

	public int getVisibleEntityCount() {
		return numberOfVisibleNpcs;
	}

}
